import PublicChannel from './channels/types/publicChannel';
import PrivateChannel from './channels/types/privateChannel';
import CustomChannel from './channels/types/customChannel';
import PublicClient from './clients/types/publicClient';
import PrivateClient from './clients/types/privateClient';
import CustomClient from './clients/types/customClient';

const logger = require('./logger');

class ChannelManager {
  channels: any[] = [];

  constructor() {
    // create default channel...
    const channel = new PublicChannel('default');
    this.channels.push(channel);
  }

  createChannel(data: any) {
    const channelExists: PublicChannel|PrivateChannel|CustomChannel|null = this.channelExists(data.channel);
    let channel: PublicChannel|PrivateChannel|CustomChannel;

    if (channelExists) {
      channel = channelExists;
      logger.accessLog.info(`Channel already exists: ${channel.id}`);
    } else {
      channel = this.createByChannelType(data);
      this.channels.push(channel);
      logger.accessLog.info(`Added channel to channel manager: ${channel.id}`);
    }

    return channel;
  }

  channelExists(channel_id: string) {
    for (const channel of this.channels) {
      if (channel.id === channel_id) {
        return channel;
      }
    }

    return null;
  }

  addClientToChannel(client: PublicClient|PrivateClient|CustomClient, channel_id: string) {
    const channel: PrivateChannel|PrivateChannel|CustomChannel|null = this.channelExists(channel_id);

    if (channel) {
      channel.addClient(client);
      client.connectToChannel(channel);
      return {status: 'success'};
    } else {
      logger.accessLog.info(`channel with id ${channel_id} does not exist.`);
      return {status: 'notice', message: `channel with id ${channel_id} does not exist.`};
    }
  }

  createByChannelType(data: any) {
    try {
      logger.accessLog.info(`attempting to create channel of type ${data.channel_type}, channel id: ${data.channel}...`);

      if (data.channel_type === 'public') {
        return new PublicChannel(data.channel);
      } else if (data.channel_type === 'private') {
        return new PrivateChannel(data.channel);
      } else {
        return new CustomChannel(data.channel, data.custom);
      }
    } catch (e) {
      console.log(e);
      logger.errorLog.info(e);
      logger.accessLog.info(`creating base channel: ${data.channel}`);
      return new PublicChannel(data.channel);
    }
  }

  removeClientFromChannel(client_id: number, channel_id: string) {
    for (const channel of this.channels) {
      if (channel.id === channel_id) {
        if (channel.removeClient(client_id)) {
          logger.accessLog.info(`client removed from channel - channel: ${channel_id}, client: ${client_id}`);
          return true;
        } else {
          logger.errorLog.info(`client not removed from channel, or doesn't exist in channel - channel: ${channel_id}, client: ${client_id}`);
          return false;
        }
      }
    }

    return false;
  }

  changeChannel(client: PublicClient|PrivateClient|CustomClient, changeRequest: any) {
    if (client.channel != null) {
      this.removeClientFromChannel(client.id, client.channel.id);
    }

    const channel = this.createChannel(changeRequest);
    this.addClientToChannel(client, channel.id);
    const message = {message_type: 'requestExamStatus'};
    channel.broadcastMessage(client, message);
    this.purgeEmptyChannels();
  }

  purgeEmptyChannels() {
    for (const channel of this.channels) {
      if (channel.clients.length === 0 && channel.id !== 'default') {
        const index = this.channels.indexOf(channel);
        this.channels.splice(index, 1);
        logger.accessLog.info(`channel removed: ${channel.id}`);
      }
    }
  }
}

export default ChannelManager;
